from abc import ABC, abstractmethod
from dataclasses import dataclass
from logging import getLogger

from isaaclab.assets import AssetBaseCfg
from isaaclab.sim.spawners.lights import DistantLightCfg, DomeLightCfg

from .mesh import CLASS_TAG, AssetMesh
from .terrain import TerrainInstance

logger = getLogger(__name__)

@dataclass
class AssetSpec(ABC):
    """A specification for a class of assets to be placed in a scene.
    SceneSpec class uses this to generate AssetInstance objects to be placed on the terrain.
    """

    name: str
    """The name of the asset. For example, "Tree" or "Rock"."""
    asset_cfg_class: type[AssetBaseCfg] = AssetBaseCfg
    """The configuration class for the asset."""

    @abstractmethod
    def generate(
        self,
        terrain: TerrainInstance,
    ) -> list["AssetInstance"]:
        """Generate instances of the asset to be placed on the terrain

        Args:
            terrain (TerrainInstance): The terrain to place the asset on

        Returns:
            list[AssetInstance]: A list of instances of the asset to be placed on the terrain
        """
        ...

    def create_instance(
        self,
        name: str,
        asset: AssetMesh,
        position: tuple[float, float, float],
        rotation: tuple[float, float, float, float],
        tags: dict[str, str] | None = None,
    ) -> "AssetInstance":
        """Create an AssetInstance object from an AssetSpec object

        Args:
            name (str): The name of the asset instance
            asset (AssetMesh): The mesh of the asset instance
            position (tuple[float, float, float]): The position of the asset instance
            rotation (tuple[float, float, float, float]): The rotation of the asset instance
            tags (dict[str, str], optional): Additional tags to add to the asset instance. Defaults to None.

        Returns:
            AssetInstance: The AssetInstance object
        """
        if tags is None:
            tags = {}
        return AssetInstance(
            self,
            asset,
            name,
            position,
            rotation,
            tags,
            asset_cfg_class=self.asset_cfg_class,
        )


class IdenticalAssetSpec(AssetSpec, ABC):
    """A specification for an asset class, that will be identical across all instances"""

    def __init__(
        self,
        name: str,
        mesh: AssetMesh,
        asset_cfg_class: type[AssetBaseCfg] = AssetBaseCfg,
        rotation: tuple[float, float, float, float] = (0, 0, 0, 1),
    ):
        """Create a new IdenticalAssetSpec object

        Args:
            name (str): The name of the asset
            mesh (AssetMesh): The mesh of the asset
            asset_cfg_class (type[AssetBaseCfg], optional): The configuration class for the asset. Defaults to AssetBaseCfg.
        """
        super().__init__(name, asset_cfg_class)
        self.mesh = mesh
        self.rotation = rotation

    @abstractmethod
    def find_positions(
        self, terrain: TerrainInstance
    ) -> list[tuple[float, float, float]]:
        """Find positions to place the asset on the terrain.
        This is essentially a devolved generate() that returns a list of positions instead of creating instances.

        Args:
            terrain (TerrainInstance): The terrain to place the asset on

        Returns:
            list[tuple[float, float, float]]: A list of positions to place the asset on the terrain
        """
        ...

    def generate(self, terrain: TerrainInstance) -> list["AssetInstance"]:
        """Generate instances of the asset to be placed on the terrain

        Internally, this method calls `self.find_positions(terrain)` to find
        positions to place the asset on the terrain, and then
        `self.create_instance(name, position, rotation)` to create instances of the asset.

        Args:
            terrain (TerrainInstance): The terrain to place the asset on

        Returns:
            list[AssetInstance]: A list of instances of the asset to be placed on the terrain
        """
        return [
            self.create_identical_instance(
                f"{self.name}_{i}",
                position,
                self.rotation,
            )
            for i, position in enumerate(self.find_positions(terrain))
        ]

    def create_identical_instance(
        self,
        name: str,
        position: tuple[float, float, float],
        rotation: tuple[float, float, float, float],
        tags: dict[str, str] | None = None,
    ) -> "AssetInstance":
        """Create an AssetInstance object from an AssetSpec object

        Args:
            name (str): The name of the asset instance
            position (tuple[float, float, float]): The position of the asset instance
            rotation (tuple[float, float, float, float]): The rotation of the asset instance
            tags (dict[str, str] | None, optional): The tags to assign to the asset instance. Defaults to None.

        Returns:
            AssetInstance: The AssetInstance object
        """
        return super().create_instance(
            name, self.mesh, position, rotation, tags
        )


class SceneAsset(ABC):
    """A scene asset that can be placed in a scene"""

    @abstractmethod
    def to_cfg(self) -> AssetBaseCfg:
        """Create a RigidObjectCfg object from an AssetInstance object

        Returns:
            AssetBaseCfg: The IsaacLab cfg object
        """
        ...

    @abstractmethod
    def get_name(self) -> str:
        """Get the name of the asset

        Returns:
            str: The name of the asset
        """
        ...


@dataclass
class AssetInstance(SceneAsset):
    """A specification for an asset to be placed in a scene"""

    asset_class: AssetSpec | None
    """The class of the asset"""
    mesh: AssetMesh
    """The mesh of the asset"""
    name: str
    """The name of the asset"""
    position: tuple[float, float, float]
    """The position of the asset"""
    rotation: tuple[float, float, float, float]
    """The rotation of the asset"""
    additional_tags: dict[str, str]
    """Additional tags for the asset"""
    asset_cfg_class: type[AssetBaseCfg] = AssetBaseCfg
    """The configuration class for the asset"""

    def to_cfg(self) -> AssetBaseCfg:
        """Create a config class object from an AssetInstance object.
        The returned type is determined by the asset_cfg_class attribute.

        Args:
            scene_name (str, optional): The name of the scene to place the asset into. Defaults to "World".

        Returns:
            asset_cfg_class: The IsaacLab cfg object
        """

        if self.asset_class is None:
            prim_path = f"/{self.name}"
        else:
            prim_path = f"/{self.asset_class.name}/{self.name}"

        obj = self.asset_cfg_class(prim_path=prim_path)

        spawner = self.mesh.to_cfg()

        if spawner.semantic_tags is None:
            spawner.semantic_tags = []

        if self.asset_class is not None:
            spawner.semantic_tags.append((CLASS_TAG, self.asset_class.name))
        spawner.semantic_tags.extend(self.additional_tags.items())

        obj.spawn = spawner

        init_state = obj.InitialStateCfg()
        init_state.pos = (self.position[0], self.position[1], self.position[2])
        init_state.rot = self.rotation
        obj.init_state = init_state

        return obj

    def get_name(self) -> str:
        """Get the name of the asset

        Returns:
            str: The name of the asset
        """
        return self.name


@dataclass
class DistantLightSpec(SceneAsset):
    """Specification for a scene distant light in which
     you can set exposure, intensity and color."""


    exposure: float = 11.0
    intensity: float = 1.0
    color: tuple[float, float, float] = (0.988, 0.957, 0.645)

    def to_cfg(self, scene_name: str = "World") -> AssetBaseCfg:

        logger.debug("Creating distant light cfg")

        light_cfg = DistantLightCfg()
        light_cfg.exposure = self.exposure
        light_cfg.intensity = self.intensity
        light_cfg.color = self.color

        state = AssetBaseCfg.InitialStateCfg()
        state.pos = (0, 0, 0)
        state.rot = (0.82294, 0.28336, -0.4656, -0.16032)

        cfg = AssetBaseCfg()
        cfg.prim_path = f"/{scene_name}/{self.get_name()}"
        cfg.spawn = light_cfg
        cfg.init_state = state
        return cfg

    def get_name(self) -> str:
        return "distant_light"

@dataclass
class DomeLightSpec(SceneAsset):
    """Specification for a scene dome light in which
     you can set exposure, intensity and color."""

    exposure: float = 0.0
    intensity: float = 1000.0
    color: tuple[float, float, float] = (0.988, 0.957, 0.645)

    def to_cfg(self, scene_name: str = "World") -> AssetBaseCfg:

        logger.debug("Creating dome light cfg")

        light_cfg = DomeLightCfg()
        light_cfg.exposure = self.exposure
        light_cfg.intensity = self.intensity
        light_cfg.color = self.color

        state = AssetBaseCfg.InitialStateCfg()
        state.pos = (0, 0, 0)

        cfg = AssetBaseCfg()
        cfg.prim_path = f"/{scene_name}/{self.get_name()}"
        cfg.spawn = light_cfg
        cfg.init_state = state
        return cfg

    def get_name(self) -> str:
        return "dome_light"



from dataclasses import dataclass
from logging import getLogger

import numpy as np
from isaaclab.assets import AssetBaseCfg
from isaaclab.sim.spawners import PreviewSurfaceCfg
from isaaclab.terrains import SubTerrainBaseCfg, TerrainGeneratorCfg

# from pxr import Sdf, UsdShade
from trimesh import Trimesh

# from .materials import MaterialHandler
from .mesh import DynamicMesh

logger = getLogger(__name__)

TERRAIN_NAME = "terrain"


@dataclass
class TerrainInstance:
    """A specification for a terrain to be placed in a scene.

    Logically, the terrain should cover the entire declared size, meaning
    holes are allowed, but that implies non-continuous terrain.

    Terrain consists of multiple meshes that are combined to form a single
    terrain. Each mesh can have specific semantic tags attached to it.
    """

    mesh: list[tuple[Trimesh, list[tuple[str, str]]]]
    """The mesh of the terrain and the tags to add to the mesh"""
    origin: tuple[float, float, float]
    """The position where the robot should spawn"""
    size: tuple[float, float]
    """The size of the terrain in meters"""
    color: tuple[float, float, float]
    """The color of the terrain"""
    material: str | None = None 
    """Path to the material mdl file"""


    def to_cfg(self) -> TerrainGeneratorCfg:
        """Create a TerrainGeneratorCfg object from a TerrainInstance object

        Please note that this method does not apply semantic tags to the terrain.

        Returns:
            TerrainGeneratorCfg: The TerrainGeneratorCfg object
        """
        logger.debug("Creating terrain generator cfg")
        sub_terrain = SubTerrainBaseCfg()
        sub_terrain.function = lambda diff, cfg: (
            [m[0] for m in self.mesh],
            np.array(self.origin),
        )
        sub_terrain.size = self.size

        terrain_cfg = TerrainGeneratorCfg()
        terrain_cfg.sub_terrains = {"main": sub_terrain}
        terrain_cfg.size = self.size

        return terrain_cfg

    def to_asset_cfg(self) -> list[AssetBaseCfg]:
        """Create a asset Cfg object from a TerrainInstance object

        Returns:
            AssetBaseCfg: The AssetBaseCfg object
        """
        logger.debug("Creating terrain asset cfg")
        res = []
        for i, (mesh, tags) in enumerate(self.mesh):
            spawner = DynamicMesh(
                mesh, self.material, PreviewSurfaceCfg(diffuse_color=self.color)
            ).to_cfg()
            spawner.semantic_tags = tags
            cfg = AssetBaseCfg(
                prim_path=f"/{TERRAIN_NAME}/{TERRAIN_NAME}_{i}",
                spawn=spawner,
            )
            cfg.init_state = cfg.InitialStateCfg()
            cfg.collision_group = -1
            res.append(cfg)
        return res



"""
STRIPE-kit is a kit of utilities wrapping over various Isaac Lab and Isaac
Sim utilities. The core idea, is that you define your scene by inheriting and
utilizing different classes provided here, ideally in a separate module, and
then after providing the task specification simply register it with gymnasium
and run your Isaac Lab RL training.

----------

Here's a step-by-step guide to get you started:

1. Create your `SceneSpec`
    This is where you define how the scene should be generated. Create a new
    class that inherits from `SceneSpec` and implement the methods as needed.
2. Define your task
    This is done via standard Isaac Lab configclasses, so you define your
    reward terms, termination terms, etc...
3. Couple your task with your scene
    This is done by creating a new instance of `TrainingSpec`, which takes in
    a `SceneSpec` and all the necessary parameters for the task.
4. Register your task with gymnasium
    `TrainingSpec` can be transformed into a `TaskEnvCfg`, which can be directly
    registered with gymnasium.
5. Run your training
    Once your task is registered, any training script that uses gymnasium works
    just fine. We actually have a convenient pre-built script for that, which
    once the package is installed should be accessible via the CLI as
    `skrl_train`.

----------

There are a few core concepts you need to understand when working with this
module.

- Spec: A specification, that defines how something should be generated
- Scene: An instance of a generated scene
- Mesh: A 3d mesh (3d model) that can be used in a scene
- configclass: The preferred Isaac Lab way of creating things, is via creating a subclass or instance of a configclass

"""

from .asset import AssetInstance, AssetSpec, IdenticalAssetSpec
from .env import TaskEnvCfg, TrainingSpec
from .factory import SceneCfgFactory
from .mesh import AssetMesh, DynamicMesh, UniversalMesh, USDMesh, instancable
from .scene_spec import SceneSpec
from .terrain import TerrainInstance

__all__ = [
    "SceneCfgFactory",
    "AssetSpec",
    "IdenticalAssetSpec",
    "TerrainInstance",
    "AssetInstance",
    "SceneSpec",
    "AssetMesh",
    "DynamicMesh",
    "USDMesh",
    "UniversalMesh",
    "TrainingSpec",
    "TaskEnvCfg",
    "instancable",
]



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

import gymnasium as gym
from isaaclab.assets import ArticulationCfg
from isaaclab.envs import ManagerBasedRLEnv, ManagerBasedRLEnvCfg, ViewerCfg
from isaaclab.sensors import SensorBaseCfg

from .factory import NFLInteractiveSceneCfg
from .scene_spec import SceneSpec


class TaskEnvCfg(ManagerBasedRLEnvCfg):
    """Configuration for a task environment, usually created by `TrainingSpec`."""

    scene: NFLInteractiveSceneCfg

    def __init__(
        self,
        scene: NFLInteractiveSceneCfg,
        viewer: ViewerCfg,
        decimation: int,
        actions: object,
        observations: object,
        events: object,
        rewards: object,
        terminations: object,
        commands: object,
        episode_length_s: float,
        sensors: Mapping[str, SensorBaseCfg],
        spec: SceneSpec,
    ) -> None:
        super().__init__(
            viewer=viewer,
            decimation=decimation,
            actions=actions,
            observations=observations,
            events=events,
            rewards=rewards,
            terminations=terminations,
            commands=commands,
            episode_length_s=episode_length_s,
        )

        self.scene = scene  # pyright: ignore[reportIncompatibleVariableOverride]
        self.sensors = sensors
        self.spec = spec

    def register(self, id: str, **kwargs: str):
        """Registers the environment within `gymnasium`.

        An instance of this class is stored within globals of this module,
        as `globals()[id]`. Gymnasium is then instructed to get this instance,
        and pass it to `NflEnvMixin`.

        Args:
            id (str): The id of the environment to register.
            **kwargs (str): Additional keyword arguments to pass to `gymnasium.register`.
        """

        globals()[id] = self
        gym.register(
            id=id,
            entry_point=f"{__name__}:NflEnvMixin",
            disable_env_checker=True,
            kwargs={"env_cfg_entry_point": f"{__name__}:{id}", **kwargs},
        )

    def __post_init__(self):
        """Post initialization."""
        # simulation settings
        self.sim.dt = 0.005


@dataclass
class TrainingSpec:
    """A complete specification for training a robot in a scene.

    Having defined your scene using a `SceneSpec`, you then have to define your
    task, by deciding which robot to use, what sensors you need, and your
    rewards, terminations, commands, etc... Consult the Isaac Lab documentation
    for more information how to do it properly.

    After creating your `TrainingSpec`, you can create a `ManagerBasedRLEnv`
    instance using the `to_env_cfg` method. This custom environment is best
    registered in gymnasium using the `register` method.
    """

    scene: SceneSpec
    robot: ArticulationCfg

    actions: object
    observations: object
    events: object

    rewards: object
    terminations: object
    commands: object

    sensors: Mapping[str, SensorBaseCfg]

    def to_env_cfg(
        self,
        view_cfg: ViewerCfg,
        decimation: int = 4,
        episode_length_s: float = 100.0,
    ) -> TaskEnvCfg:
        env_cfg = TaskEnvCfg(
            scene=NFLInteractiveSceneCfg(1, 1.0, robot=self.robot),
            viewer=view_cfg,
            decimation=decimation,
            actions=self.actions,
            observations=self.observations,
            events=self.events,
            rewards=self.rewards,
            terminations=self.terminations,
            commands=self.commands,
            episode_length_s=episode_length_s,
            sensors=self.sensors,
            spec=self.scene,
        )

        return env_cfg


class NflEnvMixin(ManagerBasedRLEnv):
    def __init__(self, cfg: TaskEnvCfg, **kwargs: Any):
        factory = cfg.spec.create_instance(cfg.scene.num_envs, cfg.scene.env_spacing)
        for name, sensor in cfg.sensors.items():
            factory.add_sensor(name, sensor)
        cfg.scene = factory.get_scene(cfg.scene.robot)
        self.terrain = factory.terrain
        super().__init__(cfg, **kwargs)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from logging import getLogger
from typing import Any

import isaacsim.core.utils.prims as prim_utils  # pyright: ignore[reportMissingImports]
import isaacsim.core.utils.semantics as semantics_utils  # pyright: ignore[reportMissingImports]
from isaaclab.sim.converters import MeshConverter, MeshConverterCfg
from isaaclab.sim.spawners import (
    MdlFileCfg,
    PreviewSurfaceCfg,
    RigidBodyMaterialCfg,
    SpawnerCfg,
    UsdFileCfg,
    VisualMaterialCfg,
    # spawn_from_mdl_file,
)
from isaaclab.terrains.utils import create_prim_from_mesh
from pxr.Usd import Prim  # pyright: ignore[reportMissingImports]
from trimesh import Trimesh, primitives

logger = getLogger(__name__)

CLASS_TAG = "class"


class AssetMesh(ABC):
    """A mesh that can be spawned within Isaac Sim"""

    @abstractmethod
    def to_cfg(self, **kwargs: Any) -> SpawnerCfg:
        """Create a SpawnerCfg object from an AssetMesh object

        Logically this should return a new object, thus nullifying any effects
        of instancing. If you wish to prevent this behavior, use the `instancable` decorator.

        Args:
            **kwargs: Additional keyword arguments, passed to the SpawnerCfg constructor

        Returns:
            SpawnerCfg: The IsaacLab cfg object
        """
        ...


@dataclass
class USDMesh(AssetMesh):
    """A mesh derived from a USD file"""

    usd_path: str

    def to_cfg(self, **kwargs: Any) -> UsdFileCfg:
        """Create a UsdFileCfg object from an AssetMesh object

        Args:
            **kwargs: Additional keyword arguments, passed to the UsdFileCfg constructor

        Returns:
            UsdFileCfg: The IsaacLab cfg object
        """
        logger.debug("Creating USD mesh cfg")
        mesh_cfg = UsdFileCfg(usd_path=self.usd_path, **kwargs)
        return mesh_cfg


class UniversalMesh(AssetMesh):
    """A mesh derived from any file format, accepted by MeshConverterCfg"""

    def __init__(self, path: str, **kwargs: Any):
        """Initialize a UniversalMesh object

        Args:
            path (str): Path to the mesh file
            kwargs: Additional keyword arguments to pass to the MeshConverterCfg
        """
        logger.debug("Creating universal mesh")
        cfg = MeshConverterCfg(path, **kwargs)
        self.converter = MeshConverter(cfg)

    def to_cfg(self, **kwargs: Any) -> UsdFileCfg:
        """Create a UsdFileCfg object utilizing the MeshConverterCfg from an AssetMesh object

        Args:
            **kwargs: Additional keyword arguments, passed to the UsdFileCfg constructor

        Returns:
            UsdFileCfg: The IsaacLab cfg object
        """
        mesh_cfg = UsdFileCfg(usd_path=self.converter.usd_path, **kwargs)
        return mesh_cfg


def instancable(cls: type[AssetMesh]) -> type[AssetMesh]:
    """Decorator to make the assets work as instancables

    Overrides `to_cfg` to return a single instance of `SpawnerCfg`
    """

    class _instancable(cls):
        spawner = None

        def to_cfg(self, **kwargs: Any) -> SpawnerCfg:
            if self.spawner is None:
                self.spawner = cls.to_cfg(self, **kwargs)
            elif kwargs:
                logger.warning("Ignoring additional keyword arguments")
            return self.spawner

    return _instancable


def apply_semantics(
    prim: Prim, type: str, value: str  # pyright: ignore[reportUnknownParameterType]
) -> None:
    """Applies a semantic type and data to a prim.

    Args:
        prim (Prim): A special type of a string that represents a prim in a stage
        type (str): Label
        value (str): Value
    """
    semantics_utils.add_update_semantics(prim, value, type)


@dataclass
class DynamicMesh(AssetMesh):
    """A dynamic mesh asset defined by a Trimesh object"""

    mesh: Trimesh
    """The Trimesh object that defines the mesh"""
    visual_material_path: str | None = None
    """Path to the visual material file, needed for MDL materials only"""
    visual_material: VisualMaterialCfg = field(default_factory=PreviewSurfaceCfg)
    """The visual material configuration for the mesh"""
    physics_material: RigidBodyMaterialCfg = field(default_factory=RigidBodyMaterialCfg)
    """The physics material configuration for the mesh"""

    # inspiration:
    # https://github.com/isaac-sim/IsaacLab/blob/963b53b96bc6140670fa0fe41d9fbafa68d8382f/source/isaaclab/isaaclab/terrains/utils.py#L61

    def to_cfg(self, **kwargs: Any) -> SpawnerCfg:
        """Converts the dynamic mesh to a SpawnerCfg object

        Args:
            **kwargs: Additional keyword arguments, passed to the SpawnerCfg constructor

        Returns:
            SpawnerCfg: A SpawnerCfg object representing the dynamic mesh
        """
        logger.debug("Creating dynamic mesh cfg")

        if self.visual_material_path:
            self.visual_material = MdlFileCfg(mdl_path=self.visual_material_path)

        def func_wrapper(  # pyright: ignore[reportUnknownParameterType]
            prim: str, cfg: SpawnerCfg, *args: Any, **kwargs: Any
        ) -> Prim:
            create_prim_from_mesh(
                prim,
                self.mesh,
                visual_material=self.visual_material,
                physics_material=self.physics_material,
                *args,
                **kwargs,
            )
            p: Prim = prim_utils.get_prim_at_path(prim)
            if cfg.semantic_tags is not None:
                for tag, value in cfg.semantic_tags:
                    apply_semantics(p, tag, value)
            return p

        return SpawnerCfg(func=func_wrapper, **kwargs)


@instancable
class DebugMesh(DynamicMesh):
    def __init__(self):
        super().__init__(mesh=primitives.Box(extents=[1.0, 1.0, 1.0]))



from copy import deepcopy
from dataclasses import MISSING
from logging import getLogger

# isaaclab imports
from isaaclab.assets import AssetBaseCfg
from isaaclab.scene import InteractiveSceneCfg
from isaaclab.sensors import SensorBaseCfg
from isaaclab.utils import configclass

from .asset import SceneAsset
from .terrain import TERRAIN_NAME, TerrainInstance

logger = getLogger(__name__)


@configclass
class NFLInteractiveSceneCfg(InteractiveSceneCfg):
    robot: AssetBaseCfg = MISSING  # pyright: ignore[reportAssignmentType]


class SceneCfgFactory:
    """A factory class for creating InteractiveSceneCfg objects from
    TerrainInstance and SceneAsset objects.

    Logically, this represents a generated scene, that for some reason
    you might want to alter or create multiple instances of.
    """

    robot_name: str = "robot"

    def __init__(
        self,
        terrain: TerrainInstance,
        num_envs: int = 1,
        env_spacing: float = 0.0,
        **kwargs: bool,
    ):
        """Create a new SceneCfgFactory object

        Args:
            num_envs (int): The number of environments to create
            env_spacing (float): The spacing between environments
        """
        self.num_envs = num_envs
        self.env_spacing = env_spacing
        self.kwargs = kwargs

        self.terrain = terrain

        self.assets: dict[str, AssetBaseCfg] = {}
        self.sensors: dict[str, SensorBaseCfg] = {}

    def add_asset(self, asset: SceneAsset) -> None:
        """Add an AssetInstance object to the factory

        Args:
            asset (AssetInstance): The AssetInstance object to add

        Raises:
            ValueError: If an asset with the same name already exists
        """
        self.assets[asset.get_name()] = asset.to_cfg()

        logger.debug(f"Added asset {asset.get_name()}")

    def add_sensor(self, name: str, sensor: SensorBaseCfg) -> None:
        """Add sensors to the scene

        Args:
            name (str): The name of the sensor
            sensor (SensorBaseCfg): The sensor configuration
        """
        self.sensors[name] = sensor

    def get_scene(
        self,
        robot: AssetBaseCfg,
    ) -> NFLInteractiveSceneCfg:
        """Gets the scene configuration

        Returns:
            NFLInteractiveSceneCfg: Shallow copy of the NFLInteractiveSceneCfg object
        """
        logger.debug("Creating scene cfg")
        robot = deepcopy(robot)
        robot.prim_path = "{ENV_REGEX_NS}/robot"
        robot.init_state.pos = self.terrain.origin

        cfg = NFLInteractiveSceneCfg(
            self.num_envs, self.env_spacing, robot=robot, **self.kwargs
        )

        for name, asset in self.assets.items():
            setattr(
                cfg,
                name,
                asset,
            )

        for name, sensor in self.sensors.items():
            setattr(
                cfg,
                name,
                sensor,
            )

        for i, asset in enumerate(self.terrain.to_asset_cfg()):
            setattr(
                cfg,
                TERRAIN_NAME + f"_{i}",
                asset,
            )

        return cfg



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from logging import getLogger

from isaaclab.assets import AssetBaseCfg

from .asset import AssetSpec, DistantLightSpec, DomeLightSpec
from .factory import SceneCfgFactory
from .mesh import DebugMesh
from .terrain import TerrainInstance

logger = getLogger(__name__)


def spawn_cfg(cfg: AssetBaseCfg) -> None:
    if cfg.spawn is not None:
        cfg.spawn.func(cfg.prim_path, cfg.spawn)
    else:
        raise ValueError(
            f"Spawn function not set for {cfg.__class__.__name__} asset"
        )


@dataclass
class SceneSpec(ABC):
    """A specification for a scene to be generated.

    You can think of this like a factory of scenes, which it is.
    Here you define how big the scenes will be, what **type** of assets they
    will have and what light should be added to the scene.

    ----------

    You **have to** implement the `generate` method, which is used to generate the
    terrain for the scene. Beyond that, for more complex scenes it's a good
    idea to implement `create_instance` on your own, but for simpler use
    cases that shouldn't be necessary.

    ----------

    The idea is, that your scene *should* have a logical palette of asset types.
    Think of it this way: a forest, will have trees, rocks, bushes etc...
    Logically you would create an AssetSpec for trees, rocks and bushes, and
    these would go to the palette. If you need contextual generation, you
    should place them in a single `AssetSpec`.

    """

    size: tuple[float, float]
    """The size of the scene"""
    palette: list[AssetSpec]
    """The palette of asset classes to be used in the scene"""
    distant_light: DistantLightSpec = field(default_factory=DistantLightSpec)
    dome_light: DomeLightSpec = field(default_factory=DomeLightSpec)
    """The light specification for the scene"""

    def add_asset(self, asset: AssetSpec):
        """Add an asset to the scene palette.

        The default implementation of `create_instance` will use that palette
        to generate instances of the assets to be placed in the scene.

        Args:
            asset (AssetSpec): The asset to add
        """
        self.palette.append(asset)

    @abstractmethod
    def generate(self) -> TerrainInstance:
        """Generate a terrain instance. This method is used to generate the
        terrain for the scene.

        While implementing this method, you can store extra data in the
        returned object, that then can be used by the asset specifications, thus
        allowing for more performant scene generation, where interesting spots
        encountered during terrain generation can be used to place assets.

        Returns:
            TerrainInstance: The generated terrain instance
        """
        ...

    def create_instance(
        self, num_envs: int = 1, env_spacing: float = 0.0, debug_models: bool = False, **kwargs: bool
    ) -> SceneCfgFactory:
        """Create a SceneCfgFactory object from the SceneSpec object.

        The default implementation, generates the terrain using the generate
        method, and then generates the assets using the asset specifications
        in the palette. The generated scene is then returned.

        Args:
            num_envs (int): The number of environments to generate
            env_spacing (float): The spacing between environments
            **kwargs: Additional keyword arguments to pass to the SceneCfgFactory

        Returns:
            SceneCfgFactory: The SceneCfgFactory object
        """
        logger.debug("Generating terrain")
        terrain = self.generate()
        factory = SceneCfgFactory(terrain, num_envs, env_spacing, **kwargs)
        for asset in self.palette:
            logger.debug(f"Generating asset {asset.name}")
            for child in asset.generate(terrain):
                if debug_models:
                    child.mesh = DebugMesh()
                factory.add_asset(child)
        logger.debug("Adding light")
        factory.add_asset(self.distant_light)
        factory.add_asset(self.dome_light)
        return factory
